unit AVIWriter;

interface

uses
  Forms,Dialogs,SysUtils,Windows,Classes,Graphics,VFW,SimpleCompressorDialogFormUnit,
  ExtendedCompressorDialogFormUnit;

type
  EAVIBuilderException=class(Exception)
  public
    constructor CreateCode(e:Cardinal;Context:string);  //Created from error code
  end;

  EUserAbortException=class(Exception);                 //User has canceled codec selection

  TAVIPixelFormat=(apfRGB16,apfRGB24,apfRGB32);

  TResamplingMethod=(rmStretch,rmCrop,rmCenter);

  TCompressorSelectionMethod=(csmNoCompression,csmAutoSelect,csmSimpleDialog,csmExtendedDialog,csmCustom);

  TCustomCompressorSelectionEvent=procedure(Sender:TObject;var SelectedIC:HIC;const Mode:Cardinal;var Quality,DataRate,FrameRate,MaxFramesBetweenKeys:Cardinal;var AutoKeyFrames:Boolean) of object;
  //when this event is called, it must return:
  //    a valid HIC in SelectedIC for the selected Mode (which can be ICMODE_COMPRESS or ICMODE_FASTCOMPRESS) or Zero if user canceled selection.
  //    Quality, DataRate, FrameRate, MaxFramesBetweenKeys and AutoKeyFrames can be set too, they are initialized with the corresponding AVIWriter values.

  TAVIWriter=class(TComponent)
  private
    FAVIPixelFormat: TAVIPixelFormat;
    FIC:HIC;
    FAVIWidth: Integer;
    FAVIHeight: Integer;
    FVideoStream:PAVIStream;
    FAVIFile:PAVIFile;
    FAVIFileName: string;
    FRecording: Boolean;

    FSourceBI:TBitmapInfo;
    FDestBI,FLastDestBI:PBitmapInfo;
    FSourceBits,FDestBits,FLastSourceBits:Pointer;

    FICInfo:TICInfo;
    FOnStopRecording: TNotifyEvent;
    FOnFrameAdd: TNotifyEvent;
    FOnStartRecording: TNotifyEvent;
    FResamplingMethod: TResamplingMethod;
    FAVIRealTime: Boolean;
    FCompressorSelection: TCompressorSelectionMethod;
    FOnCustomSelectCompressor: TCustomCompressorSelectionEvent;

    FDataBitmap,FSamplingBitmap:TBitmap;
    FAVIBackGroundColor: TColor;
    FAVIQuality: Cardinal;
    FSamplesWritten: LongInt;
    FBytesWritten: LongInt;
    FCurrentFrame,FLastKeyFrame:Integer;
    FAVIFrameRate: Cardinal;
    FAVIDataRate: Cardinal;
    FAVIAutoKeyFrames: Boolean;
    FAVIMaxFramesBetweenKeys: Cardinal;

    procedure SetAVIPixelFormat(const Value: TAVIPixelFormat);
    procedure SetAVIHeight(const Value: Integer);
    procedure SetAVIWidth(const Value: Integer);
    procedure SetAVIFileName(const Value: string);
    procedure SetOnFrameAdd(const Value: TNotifyEvent);
    procedure SetOnStartRecording(const Value: TNotifyEvent);
    procedure SetOnStopRecording(const Value: TNotifyEvent);
    procedure SetResamplingMethod(const Value: TResamplingMethod);
    procedure SetAVIRealTime(const Value: Boolean);
    procedure SetCompressorSelection(
      const Value: TCompressorSelectionMethod);
    procedure SetOnCustomSelectCompressor(
      const Value: TCustomCompressorSelectionEvent);
    procedure SetAVIBackGroundColor(const Value: TColor);
    procedure SetAVIQuality(const Value: Cardinal);

    procedure AllocateDataBitmap(BackGround:TColor);
    procedure AllocateSamplingBitmap(AWidth,AHeight:Integer);
    procedure ExtractDataBitmapBits;
    procedure SetAVIDataRate(const Value: Cardinal);
    procedure SetAVIFrameRate(const Value: Cardinal);
    procedure SetAVIAutoKeyFrames(const Value: Boolean);
    procedure SetAVIMaxFramesBetweenKeys(const Value: Cardinal);
  protected
    function GetBitCount(f:TAVIPixelFormat):Word;

    procedure InitializeCompress;
    procedure FinalizeCompress;
    procedure AddFrame;

    procedure PrepareIC;

    function ICNeedDoubleBuffer:Boolean;
    function ICNeedQuality:Boolean;
    function ICNeedFrameSize:Boolean;
  public
    constructor Create(AOwner:TComponent);override;

    procedure StartRecording;                                       //Start AVI recording (no processing parameters can change from now on)
    procedure StopRecording;                                        //Stop AVI recording and finalize file. MUST be called after each call to StartRecording,
                                                                    //or the generated file will be corrupted.

    procedure AddFrameBitmap(ABitmap:TBitmap);                      //Add a bitmap image to AVI stream
    procedure AddFrameGraphic(AGraphic:TGraphic);                   //Add graphic object (ie: TJPEGImage) to stream
    procedure AddFrameFromDC(SrcDC:HDC;SrcRect:TRect);              //Add frame from Windows device context (ie: GetDC(GetDesktopWindow)).
                                                                    //SrcRect is the source image rectangle.

    procedure AddFrameRawRGB(RGBPtr:Pointer;Width,Height:Integer);  //Add RGB data pointer. Data must be 32 bits-aligned.

    property BytesWritten:LongInt read FBytesWritten;               //Number of bytes written so far
    property SamplesWritten:LongInt read FSamplesWritten;           //Number of "samples" written so far (used by AVIStreamWrite, seems buggy, so do not use...)
    property CurrentFrame:Integer read FCurrentFrame;               //Current frame index

    destructor Destroy;override;
  published
    //The following properties MUST be set BEFORE StartRecording:
    property AVIPixelFormat:TAVIPixelFormat read FAVIPixelFormat write SetAVIPixelFormat default apfRGB24;           //AVI pixel format, default RGB24.
    property AVIFileName:string read FAVIFileName write SetAVIFileName;                                              //AVI file name.
    property AVIWidth:Integer read FAVIWidth write SetAVIWidth;                                                      //AVI pixel width.
    property AVIHeight:Integer read FAVIHeight write SetAVIHeight;                                                   //AVI pixel height.
    property AVIRealTime:Boolean read FAVIRealTime write SetAVIRealTime;                                             //Fast compress mode.
    property AVIFrameRate:Cardinal read FAVIFrameRate write SetAVIFrameRate;                                         //Frame rate (number of frames in 1000 seconds)

    //The following properties can be set whil recording:
    property AVIBackGroundColor:TColor read FAVIBackGroundColor write SetAVIBackGroundColor default clNone;          //Background color for stretching.
    property AVIQuality:Cardinal read FAVIQuality write SetAVIQuality;                                               //Quality. Not all codecs can support setting quality values.
    property AVIDataRate:Cardinal read FAVIDataRate write SetAVIDataRate;                                            //Dat arate. Not all codecs can support setting data rate values.
    property AVIAutoKeyFrames:Boolean read FAVIAutoKeyFrames write SetAVIAutoKeyFrames;                              //Auto-generate keyframes. Some codecs make that themselves.
    property AVIMaxFramesBetweenKeys:Cardinal read FAVIMaxFramesBetweenKeys write SetAVIMaxFramesBetweenKeys;        //Maximum number of frames between 2 keyframes. Used only if AutoKeyFrame=True.

    property ResamplingMethod:TResamplingMethod read FResamplingMethod write SetResamplingMethod;                    //Sampling method when added source image has not the same size than the AVI.
    property CompressorSelection:TCompressorSelectionMethod read FCompressorSelection write SetCompressorSelection;  //Compression selection method:
                                                                                                                     // * csmNoCompression:  no compression
                                                                                                                     // * csmAutoSelect:     let windows select appropriate codec (not recommended! lol)
                                                                                                                     // * csmSimpleDialog:   simple dialog selection
                                                                                                                     // * csmExtendedDialog: extended dialog
                                                                                                                     // * csmCustom:         use OnCustomSelectCompressor event, which must be set or an EUserAbortException will be raised

    property OnStartRecording:TNotifyEvent read FOnStartRecording write SetOnStartRecording;  //Recording has successfully started
    property OnStopRecording:TNotifyEvent read FOnStopRecording write SetOnStopRecording;     //Recording is being stopped
    property OnFrameAdd:TNotifyEvent read FOnFrameAdd write SetOnFrameAdd;                    //a frame is being added

    property OnCustomSelectCompressor:TCustomCompressorSelectionEvent read FOnCustomSelectCompressor write SetOnCustomSelectCompressor;  //see CompressorSelection and TCustomCompressorSelectionEvent
  end;

const
  GIC_NOCOMPRESS=Cardinal(-1);  //custom HIC for "no compression"

procedure AVICheck(e:Cardinal;Context:string='');
function Align32(x:Integer):Integer;
procedure WinGetMem(var p;Size:DWord);
procedure WinFreeMem(p:Pointer);

implementation

uses Math;

{$R AVIWriter.dcr}

procedure AVICheck(e:Cardinal;Context:string);
begin
  if e<>AVIERR_OK then
    raise EAVIBuilderException.CreateCode(e,Context);
end;

function Align32(x:Integer):Integer;
begin
  if x mod 32=0 then
    Result:=(x div 32)*4
  else
    Result:=4+4*(x div 32);
end;

procedure WinGetMem(var p;Size:Cardinal);
begin
  Pointer(p):=VirtualAlloc(nil,Size,MEM_COMMIT or MEM_RESERVE,PAGE_READWRITE);
  if not Assigned(Pointer(p)) then
    RaiseLastOSError;
end;

procedure WinFreeMem(p:Pointer);
begin
  if Assigned(p) and not VirtualFree(p,0,MEM_RELEASE) then
    RaiseLastOSError;
end;

{ EAVIBuilderException }

constructor EAVIBuilderException.CreateCode(e: Cardinal; Context: string);
var
  s:string;
begin
  case e of
    AVIERR_UNSUPPORTED:s:='Unsupported';
    AVIERR_BADFORMAT:s:='Bad format';
    AVIERR_MEMORY:s:='Out of memory';
    AVIERR_INTERNAL:s:='Internal error';
    AVIERR_BADFLAGS:s:='Bad flags';
    AVIERR_BADPARAM:s:='Bad parameter';
    AVIERR_BADSIZE:s:='Bad size';
    AVIERR_BADHANDLE:s:='Bad handle';
    AVIERR_FILEREAD:s:='File read error';
    AVIERR_FILEWRITE:s:='File write error';
    AVIERR_FILEOPEN:s:='File already open';
    AVIERR_COMPRESSOR:s:='Compressor error';
    AVIERR_NOCOMPRESSOR:s:='No compressor';
    AVIERR_READONLY:s:='File is read-only';
    AVIERR_NODATA:s:='No data';
    AVIERR_BUFFERTOOSMALL:s:='Buffer too small';
    AVIERR_CANTCOMPRESS:s:='Can''t compress';
    AVIERR_USERABORT:s:='User abort';
//    AVIERR_ERROR:s:='Error';
  else
    s:=SysErrorMessage(e);//'Unspecified error';
    if s='' then
      s:=Format('Unspecified error (%x)',[e]);
  end;
  if Context<>'' then
    inherited Create(Context+#13+s)
  else
    inherited Create(s);
end;

{ TAVIWriter }

procedure TAVIWriter.AddFrame;
var
  Flag,Quality,FrameSize:Cardinal;
  p:Pointer;
  SamplesWritten,BytesWritten:Longint;
begin
  if Assigned(FOnFrameAdd) then
    FOnFrameAdd(Self);
  if FIC<>GIC_NOCOMPRESS then begin
    if FAVIAutoKeyFrames and ((FCurrentFrame-FLastKeyFrame) mod Integer(1+FAVIMaxFramesBetweenKeys)=0) then
      Flag:=ICCOMPRESS_KEYFRAME
    else
      Flag:=0;
    if ICNeedQuality then
      Quality:=FAVIQuality
    else
      Quality:=0;
    if ICNeedFrameSize then
      FrameSize:=(1000*FAVIDataRate) div FAVIFrameRate
    else
      FrameSize:=0;
    if (Flag<>0) or not ICNeedDoubleBuffer then
      AVICheck(ICCompress(FIC,Flag,@FDestBI.BMIHeader,FDestBits,@FSourceBI.BMIHeader,FSourceBits,nil,@Flag,FCurrentFrame,FrameSize,Quality,nil,nil),'Cannot compress frame #'+IntToStr(FCurrentFrame))
    else
      AVICheck(ICCompress(FIC,Flag,@FDestBI.BMIHeader,FDestBits,@FSourceBI.BMIHeader,FSourceBits,nil,@Flag,FCurrentFrame,FrameSize,Quality,@FLastDestBI.BMIHeader,FLastSourceBits),'Cannot compress frame #'+IntToStr(FCurrentFrame));
    if ICNeedDoubleBuffer then begin
      p:=FDestBI;
      FDestBI:=FLastDestBI;
      FLastDestBI:=p;
      p:=FSourceBits;
      FSourceBits:=FLastSourceBits;
      FLastSourceBits:=p;
    end;
  end else
    Flag:=AVIIF_KEYFRAME;
  if Flag<>0 then
    FLastKeyFrame:=FCurrentFrame;
  Inc(FCurrentFrame);
  if FIC=GIC_NOCOMPRESS then
    AVICheck(AVIStreamWrite(FVideoStream,FCurrentFrame,1,FSourceBits,Align32(FAVIWidth*GetBitCount(FAVIPixelFormat))*FAVIHeight,Flag,@SamplesWritten,@BytesWritten),'Cannot write frame #'+IntToStr(FCurrentFrame)+' to stream')
  else
    AVICheck(AVIStreamWrite(FVideoStream,FCurrentFrame,1,FDestBits,FDestBI.BMIHeader.biSizeImage,Flag,@SamplesWritten,@BytesWritten),'Cannot write frame #'+IntToStr(FCurrentFrame)+' to stream');
  FSamplesWritten:=FSamplesWritten+SamplesWritten;
  FBytesWritten:=FBytesWritten+BytesWritten;
end;

procedure TAVIWriter.AddFrameBitmap(ABitmap: TBitmap);
begin
  if (ABitmap.Width=FAVIWidth) and (ABitmap.Height=FAVIHeight) then begin
    if GetDIBits(ABitmap.Canvas.Handle,ABitmap.Handle,0,FAVIHeight,FSourceBits,FSourceBI,DIB_RGB_COLORS)=0 then
      RaiseLastOSError;
  end else begin
    AllocateDataBitmap(ABitmap.Canvas.Pixels[0,0]);
    case FResamplingMethod of
      rmStretch:StretchBlt(FDataBitmap.Canvas.Handle,0,0,FAVIWidth,FAVIHeight,ABitmap.Canvas.Handle,0,0,ABitmap.Width,ABitmap.Height,SRCCOPY);
      rmCrop:FDataBitmap.Canvas.Draw(0,0,ABitmap);
      rmCenter:FDataBitmap.Canvas.Draw((FAVIWidth-ABitmap.Width) div 2,(FAVIHeight-ABitmap.Height) div 2,ABitmap);
    end;
    ExtractDataBitmapBits;
  end;
  AddFrame;
end;

procedure TAVIWriter.AddFrameFromDC(SrcDC: HDC; SrcRect: TRect);
begin
  with SrcRect do begin
    AllocateDataBitmap(GetPixel(SrcDC,SrcRect.Left,srcRect.Top));
    if (Right-Left=FAVIWidth) and (Bottom-Top=FAVIHeight) then 
      BitBlt(FDataBitmap.Canvas.Handle,0,0,FAVIWidth,FAVIHeight,SrcDC,Left,Top,SRCCOPY)
    else
      case FResamplingMethod of
        rmStretch:StretchBlt(FDataBitmap.Canvas.Handle,0,0,FAVIWidth,FAVIHeight,SrcDC,Left,Top,Right-Left,Bottom-Top,SRCCOPY);
        rmCrop:BitBlt(FDataBitmap.Canvas.Handle,0,0,Right-Left,Bottom-Top,SrcDC,Left,Top,SRCCOPY);
        rmCenter:BitBlt(FDataBitmap.Canvas.Handle,(FAVIWidth-Right+Left) div 2,(FAVIHeight-Bottom+Top) div 2,Right-Left,Bottom-Top,SrcDC,Left,Top,SRCCOPY);
      end;
  end;
  ExtractDataBitmapBits;
  AddFrame;
end;

procedure TAVIWriter.AddFrameGraphic(AGraphic: TGraphic);
begin
  if AGraphic is TBitmap then
    AddFrameBitmap(TBitmap(AGraphic))
  else begin
    AllocateDataBitmap(0);
    AllocateSamplingBitmap(1,1);
    AllocateDataBitmap(FSamplingBitmap.Canvas.Pixels[0,0]);
    if (AGraphic.Width=FAVIWidth) and (AGraphic.Height=FAVIHeight) then
      FDataBitmap.Canvas.Draw(0,0,AGraphic)
    else begin
      FSamplingBitmap.Canvas.Draw(0,0,AGraphic);
      case FResamplingMethod of
        rmStretch:FDataBitmap.Canvas.StretchDraw(Rect(0,0,FAVIWidth,FAVIheight),AGraphic);
        rmCrop:FDataBitmap.Canvas.Draw(0,0,AGraphic);
        rmCenter:FDataBitmap.Canvas.Draw((FAVIWidth-AGraphic.Width) div 2,(FAVIHeight-AGraphic.Height) div 2,AGraphic);
      end;
    end;
    ExtractDataBitmapBits;
    AddFrame;
  end;
end;

procedure TAVIWriter.AddFrameRawRGB(RGBPtr: Pointer; Width,
  Height: Integer);
var
  BI:TBitmapInfo;
  p:PByteArray;
begin
  if (GetBitCount(FAVIPixelFormat)=24) and (Width=FAVIWidth) and (Height=FAVIHeight) then
    CopyMemory(FSourceBits,RGBPtr,Align32(Width*24)*Height)
  else begin
    p:=PByteArray(RGBPtr);
    if (Width=0) or (Height=0) then
      AllocateDataBitmap(clNone)
    else begin
      AllocateDataBitmap(RGB(p[0],p[1],p[2]));
      ZeroMemory(@BI,SizeOf(BI));
      with BI.bmiHeader do begin
        biSize:=SizeOf(TBitmapInfoHeader);
        biWidth:=Width;
        biHeight:=Height;
        biPlanes:=1;
        biBitCount:=24;
        biCompression:=BI_RGB;
      end;
      case FResamplingMethod of
        rmStretch:StretchDIBits(FDataBitmap.Canvas.Handle,0,0,FAVIWidth,FAVIHeight,0,0,Width,Height,RGBPtr,BI,DIB_RGB_COLORS,SRCCOPY);
        rmCrop:StretchDIBits(FDataBitmap.Canvas.Handle,0,0,FAVIWidth,FAVIHeight,0,0,FAVIWidth,FAVIHeight,RGBPtr,BI,DIB_RGB_COLORS,SRCCOPY);
        rmCenter:StretchDIBits(FDataBitmap.Canvas.Handle,(FAVIWidth-Width) div 2,(FAVIHeight-Height) div 2,Width,Height,0,0,Width,Height,RGBPtr,BI,DIB_RGB_COLORS,SRCCOPY);
      end;
    end;
    ExtractDataBitmapBits;
  end;
  AddFrame;
end;

procedure TAVIWriter.AllocateDataBitmap(BackGround: TColor);
begin
  if not Assigned(FDataBitmap) then begin
    FDataBitmap:=TBitmap.Create;
    FDataBitmap.PixelFormat:=pf24bit;
    FDataBitmap.Width:=FAVIWidth;
    FDataBitmap.Height:=FAVIHeight;
  end;
  if FResamplingMethod=rmStretch then
    SetStretchBltMode(FDataBitmap.Canvas.Handle,STRETCH_HALFTONE);
  if BackGround<>clNone then begin
    if FAVIBackGroundColor=clNone then
      FDataBitmap.Canvas.Brush.Color:=BackGround
    else
      FDataBitmap.Canvas.Brush.Color:=FAVIBackGroundColor;
  end;
  FDataBitmap.Canvas.FillRect(Rect(0,0,FAVIWidth,FAVIHeight));
end;

procedure TAVIWriter.AllocateSamplingBitmap(AWidth, AHeight: Integer);
begin
  if not Assigned(FSamplingBitmap) then begin
    FSamplingBitmap:=TBitmap.Create;
    FDataBitmap.PixelFormat:=pf24bit;
  end;
  if FSamplingBitmap.Width<>AWidth then
    FSamplingBitmap.Width:=AWidth;
  if FSamplingBitmap.Height<>AHeight then
    FSamplingBitmap.Height:=AHeight;
end;

constructor TAVIWriter.Create(AOwner: TComponent);
begin
  inherited;
  AVIFileInit;
  FAVIPixelFormat:=apfRGB24;
  FAVIWidth:=160;
  FAVIHeight:=120;
  FAVIBackGroundColor:=clNone;
  FAVIQuality:=7000;
  FAVIAutoKeyFrames:=False;
  FAVIMaxFramesBetweenKeys:=9;
  FAVIFrameRate:=30000;
  FAVIDataRate:=300000;
  FCompressorSelection:=csmAutoSelect;
end;

destructor TAVIWriter.Destroy;
begin
  AVIFileExit;
  inherited;
end;

procedure TAVIWriter.ExtractDataBitmapBits;
begin
  if GetDIBits(FDataBitmap.Canvas.Handle,FDataBitmap.Handle,0,FAVIHeight,FSourceBits,FSourceBI,DIB_RGB_COLORS)=0 then
    RaiseLastOSError;
end;

procedure TAVIWriter.FinalizeCompress;

  procedure FreePtr(var p:Pointer);
  begin
    if Assigned(p) then begin
      WinFreeMem(p);
      p:=nil;
    end;
  end;

begin
  if FRecording then begin
    FRecording:=False;
    if (FIC<>0) and (FIC<>GIC_NOCOMPRESS) then
      AVICheck(ICCompressEnd(FIC),'Cannot finalize compression');
  end;
  if Assigned(FVideoStream) then begin
    AVICheck(AviStreamRelease(FVideoStream),'Cannot release stream');
    FVideoStream:=nil;
  end;
  if Assigned(FAVIFile) then begin
    AVICheck(AVIFileRelease(FAVIFile),'Cannot close file');
    FAVIFile:=nil;
  end;
  if FIC<>0 then begin
    if FIC<>GIC_NOCOMPRESS then
      AVICheck(ICClose(FIC),'Cannot close compressor codec');
    FIC:=0;
  end;
  FreePtr(Pointer(FDestBI));
  FreePtr(Pointer(FLastDestBI));
  FreePtr(FSourceBits);
  FreePtr(FLastSourceBits);
  FreePtr(FDestBits);
end;

function TAVIWriter.GetBitCount(f: TAVIPixelFormat): Word;
const
  T:array[apfRGB16..apfRGB32] of Word=(16,24,32);
begin
  Result:=T[f];
end;

function TAVIWriter.ICNeedDoubleBuffer: Boolean;
begin
  if FAVIRealTime or (FIC=GIC_NOCOMPRESS) then
    Result:=True
  else
    Result:=(FICInfo.dwFlags and VIDCF_TEMPORAL=VIDCF_TEMPORAL) and not (FICInfo.dwFlags and VIDCF_FASTTEMPORALC=VIDCF_FASTTEMPORALC);
end;

function TAVIWriter.ICNeedFrameSize: Boolean;
begin
  Result:=(FIC<>GIC_NOCOMPRESS) and (FICInfo.dwFlags and VIDCF_CRUNCH=VIDCF_CRUNCH);
end;

function TAVIWriter.ICNeedQuality: Boolean;
begin
  Result:=(FIC<>GIC_NOCOMPRESS) and (FICInfo.dwFlags and VIDCF_QUALITY=VIDCF_QUALITY);
end;

procedure TAVIWriter.InitializeCompress;
var
  SI:TAVIStreamInfo;
begin
  try
    ZeroMemory(@FSourceBI,SizeOf(TBitmapInfo));
    with FSourceBI.bmiHeader do begin
      biSize:=SizeOf(TBitmapInfoHeader);
      biWidth:=FAVIWidth;
      biHeight:=FAVIHeight;
      biPlanes:=1;
      biBitCount:=GetBitCount(FAVIPixelFormat);
      biCompression:=BI_RGB;
      biSizeImage:=Align32(FAVIWidth*biBitCount)*FAVIHeight;
    end;

    PrepareIC;

    if FIC<>GIC_NOCOMPRESS then begin
      if ICGetInfo(FIC,@FICInfo,SizeOf(TICInfo))=0 then
        raise EAVIBuilderException.Create('Unable to query compressor info');
      WinGetMem(FDestBI,ICCompressGetFormatSize(FIC,@FSourceBI.bmiHeader));
      AVICheck(ICCompressGetFormat(FIC,@FSourceBI.bmiHeader,@FDestBI.bmiHeader),'Cannot get output format from codec');
      WinGetMem(FDestBits,ICCompressGetSize(FIC,@FSourceBI.BMIHeader,@FDestBI.BMIHeader));
      if ICNeedDoubleBuffer then begin
        WinGetMem(FLastDestBI,ICCompressGetFormatSize(FIC,@FSourceBI.bmiHeader));
        AVICheck(ICCompressGetFormat(FIC,@FSourceBI.bmiHeader,@FLastDestBI.bmiHeader),'Cannot get output format from codec');
      end;
      AVICheck(ICCompressQuery(FIC,@FSourceBI.BMIHeader,@FDestBI.BMIHeader),'Codec does not support format');
    end else begin
      WinGetMem(FDestBI,SizeOf(TBitmapInfo));
      CopyMemory(FDestBI,@FSourceBI,SizeOf(TBitmapInfoHeader));
    end;

    if FileExists(FAVIFileName) then
      DeleteFile(PChar(FAVIFileName));
    AVICheck(AVIFileOpen(FAVIFile,PChar(FAVIFileName),OF_WRITE or OF_CREATE OR OF_SHARE_EXCLUSIVE,nil),'Cannot open file "'+FAVIFileName+'"');

    FillChar(SI,SizeOf(SI),0);

    with SI do begin
      dwRate:=FAVIFrameRate;
      dwScale:=1000;
      fccType:=streamtypeVIDEO;
      if FIC<>GIC_NOCOMPRESS then
        fccHandler:=FICInfo.fccHandler
      else
        fccHandler:=0;
      rcFrame.Right:=FAVIWidth;
      rcFrame.Bottom:=FAVIHeight;
    end;

    AVICheck(AVIFileCreateStream(FAVIFile,FVideoStream,@SI),'Cannot create stream');

//    showmessage(inttostr(FSourceBI.bmiHeader.biBitCount)+' --> '+inttostr(FDestBI.bmiHeader.biBitCount));

    AVICheck(AVIStreamSetFormat(FVideoStream,0,@FDestBI.bmiHeader,FDestBI.bmiHeader.biSize),'Cannot set AVI stream format');

    WinGetMem(FSourceBits,Align32(FAVIWidth*FSourceBI.bmiHeader.biBitCount)*FAVIHeight);
    if ICNeedDoubleBuffer then
      WinGetMem(FLastSourceBits,Align32(FAVIWidth*FSourceBI.bmiHeader.biBitCount)*FAVIHeight);

    if FIC<>GIC_NOCOMPRESS then
      AVICheck(ICCompressBegin(FIC,@FSourceBI.BMIHeader,@FDestBI.BMIHeader),'Cannot initialize compression');


    FRecording:=True;
  except
    FinalizeCompress;
    raise;
  end;
end;

procedure TAVIWriter.PrepareIC;

  function Mode:Cardinal;
  begin
    if AVIRealTime then
      Result:=ICMODE_COMPRESS
    else
      Result:=ICMODE_FASTCOMPRESS;
  end;

begin
  FIC:=0;
  case FCompressorSelection of
    csmNoCompression:FIC:=GIC_NOCOMPRESS;
    csmAutoSelect:FIC:=ICLocate(ICTYPE_VIDEO,0,@FSourceBI.BMIHeader,nil,Mode);
    csmSimpleDialog:FIC:=ExecuteSimpleCompressorDialog(FSourceBI.bmiHeader,Mode,FAVIQuality,FAVIDataRate,FAVIFrameRate,FAVIMaxFramesBetweenKeys,FAVIAutoKeyFrames);
    csmExtendedDialog:FIC:=ExecuteExtendedCompressorDialog(FSourceBI.bmiHeader,Mode,FAVIQuality,FAVIDataRate,FAVIFrameRate,FAVIMaxFramesBetweenKeys,FAVIAutoKeyFrames);
    csmCustom:if Assigned(FOnCustomSelectCompressor) then
                FOnCustomSelectCompressor(Self,FIC,Mode,FAVIQuality,FAVIDataRate,FAVIFrameRate,FAVIMaxFramesBetweenKeys,FAVIAutoKeyFrames);
  end;
  if FIC=0 then
    raise EUserAbortException.Create('Aborted by user');
end;

procedure TAVIWriter.SetAVIAutoKeyFrames(const Value: Boolean);
begin
  FAVIAutoKeyFrames := Value;
end;

procedure TAVIWriter.SetAVIBackGroundColor(const Value: TColor);
begin
  FAVIBackGroundColor := Value;
end;

procedure TAVIWriter.SetAVIDataRate(const Value: Cardinal);
begin
  FAVIDataRate := Value;
end;

procedure TAVIWriter.SetAVIFileName(const Value: string);
begin
  Assert(not Frecording,'AVI file name be changed while recording');
  FAVIFileName := Value;
end;

procedure TAVIWriter.SetAVIFrameRate(const Value: Cardinal);
begin
  Assert(not Frecording,'AVI frame rate cannot be changed while recording');
  FAVIFrameRate := Value;
end;

procedure TAVIWriter.SetAVIHeight(const Value: Integer);
begin
  Assert(not Frecording,'AVI size cannot be changed while recording');
  FAVIHeight := Value;
end;

procedure TAVIWriter.SetAVIMaxFramesBetweenKeys(const Value: Cardinal);
begin
  FAVIMaxFramesBetweenKeys := Value;
end;

procedure TAVIWriter.SetAVIPixelFormat(const Value: TAVIPixelFormat);
begin
  Assert(not Frecording,'Pixel format cannot be changed while recording');
  FAVIPixelFormat := Value;
end;

procedure TAVIWriter.SetAVIQuality(const Value: Cardinal);
begin
  FAVIQuality := Value;
end;

procedure TAVIWriter.SetAVIRealTime(const Value: Boolean);
begin
  Assert(not Frecording,'Real time property cannot be set while recording');
  FAVIRealTime := Value;
end;

procedure TAVIWriter.SetAVIWidth(const Value: Integer);
begin
  Assert(not Frecording,'AVI size cannot be changed while recording');
  FAVIWidth := Value;
end;

procedure TAVIWriter.SetCompressorSelection(
  const Value: TCompressorSelectionMethod);
begin
  FCompressorSelection := Value;
end;

procedure TAVIWriter.SetOnCustomSelectCompressor(
  const Value: TCustomCompressorSelectionEvent);
begin
  FOnCustomSelectCompressor := Value;
end;

procedure TAVIWriter.SetOnFrameAdd(const Value: TNotifyEvent);
begin
  FOnFrameAdd := Value;
end;

procedure TAVIWriter.SetOnStartRecording(const Value: TNotifyEvent);
begin
  FOnStartRecording := Value;
end;

procedure TAVIWriter.SetOnStopRecording(const Value: TNotifyEvent);
begin
  FOnStopRecording := Value;
end;

procedure TAVIWriter.SetResamplingMethod(const Value: TResamplingMethod);
begin
  FResamplingMethod := Value;
end;

procedure TAVIWriter.StartRecording;
begin
  Assert(FAVIWidth<>0,'AVI width cannot be zero');
  Assert(FAVIHeight<>0,'AVI height cannot be zero');
  Assert(not FRecording,'Already recording!');
  InitializeCompress;
  FSamplesWritten:=0;
  FBytesWritten:=0;
  FCurrentFrame:=0;
  FLastKeyFrame:=0;
  if Assigned(FOnStartRecording) then
    FOnStartRecording(Self);
end;

procedure TAVIWriter.StopRecording;
begin
  Assert(FRecording,'Recording not launched yet!');
  if Assigned(FOnStopRecording) then
    FOnStopRecording(Self);
  if Assigned(FDataBitmap) then
    FreeAndNil(FDataBitmap);
  if Assigned(FSamplingBitmap) then
    FreeAndNil(FSamplingBitmap);
  FinalizeCompress;
end;

end.
